package org.naokishibata.armarker;

import java.util.*;
import java.io.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.imageio.*;

import org.naokishibata.generic.*;
import org.naokishibata.math.*;
import org.naokishibata.sleef.*;
import org.naokishibata.ocvimgio.*;

import static org.jocl.CL.*;
import org.jocl.*;

public class Calib {
    class Node {
	LinkedHashSet<Node> children = new LinkedHashSet<Node>();
	final Vector2D ul, lr;
	final Vector2D[] vt;
	final int ptr;
	final double ar;

	int[] nn;
	Node[] cn;

	Node(int ptr, Vector2D[] vt, Vector2D ul, Vector2D lr) {
	    this.ptr = ptr;
	    this.vt = vt;
	    this.ul = ul;
	    this.lr = lr;
	    this.ar = (lr.x - ul.x) / (lr.y - ul.y + 1e-100);
	}

	public String toString() {
	    return "(" + ul.x + "," + ul.y + "-" + lr.x + "," + lr.y + ")";
	}
    }

    static void drawX(Scribble2D s, Vector3D v, Color c) {
	s.drawLine(v.x-5, v.y-5, v.x+5, v.y+5, 2, c);
	s.drawLine(v.x-5, v.y+5, v.x+5, v.y-5, 2, c);
    }

    static void drawX(Scribble2D s, Vector2D v, Color c) {
	s.drawLine(v.x-5, v.y-5, v.x+5, v.y+5, 2, c);
	s.drawLine(v.x-5, v.y+5, v.x+5, v.y-5, 2, c);
    }

    JFrame frame;
    Scribble2D scr;

    IplImage iimg;

    BufferedImage image;
    int iw, ih;
    double[][] buf;
    int[][] ibuf;
    byte[] imgBuf;
    int[] rgbBuf;

    boolean benchMode = false;
    boolean enable_pass9 = false;
    boolean skip_cpu_part = false;

    //

    double aovid = 55.0; // logitech C910
    //double aovid = 53.0; // maxell
    double aov = aovid / 180.0 * Math.PI;
    double tanaov = Math.tan(aov / 2);
    final int SHIFT = 11;
    final int MAPSIZE = 1 << SHIFT;
    final int EDGE = 1;

    final int XNUMSTEP = 8, YNUMSTEP = 32;

    final int modeThreshold = 120;

    CLHelper cls;
    cl_mem clSrcBuf2, clDstBuf2, clSrcImage;

    Color[] cols = {Color.red, Color.green, Color.blue, Color.cyan, Color.magenta, Color.yellow, Color.pink, Color.black};

    Scribble2D s0 = null;

    //

    //NMInst nmInst = new NMInst();

    //

    cl_kernel makeKernel(String fn, String kName, int iw, int ih) throws Exception {
	String s = null;

	System.out.println("compiling " + fn);

	s = MiscUtils.readWholeFileAsAString(fn);

	s = s.replace("%KERNEL_NAME%", kName);
	s = s.replace("%WIDTH%", "" + iw);
	s = s.replace("%HEIGHT%", "" + ih);
	s = s.replace("%SHIFT%", "" + SHIFT);

	s = s.replace("%XNUMSTEP%", "" + XNUMSTEP);
	s = s.replace("%YNUMSTEP%", "" + YNUMSTEP);

	cl_kernel k = cls.compile(kName, s, null);

	return k;
    }

    static double cubicSub(double p0, double p1, double p2, double p3, double x) {
	double u, v, w;
	v = p1 - p2;
	w = p3 - p0;
	u = v * 3.0 + w;
	u = u * x   + (-4.0 * v + (p0 - p1 - w));
	u = u * x   + (p2 - p0);
	u = u * x * 0.5 + p1;
	return u;
    }

    int pixrgb(int y, int x) {
	return (rgbBuf[(y + 0) * MAPSIZE + (x + 0)] >> 8) & 255;
    }

    double bicubic(double vx, double vy) {
	if (vy < 1 || vy >= ih-2 || vx < 1 || vx >= iw-2) return 0;

	int ix = (int)vx;
	double x = vx - ix;
	int iy = (int)vy;
	double y = vy - iy;
	
	return cubicSub(cubicSub(pixrgb(iy-1, ix-1), pixrgb(iy-1, ix), pixrgb(iy-1, ix+1), pixrgb(iy-1, ix+2), x),
			cubicSub(pixrgb(iy+0, ix-1), pixrgb(iy+0, ix), pixrgb(iy+0, ix+1), pixrgb(iy+0, ix+2), x),
			cubicSub(pixrgb(iy+1, ix-1), pixrgb(iy+1, ix), pixrgb(iy+1, ix+1), pixrgb(iy+1, ix+2), x),
			cubicSub(pixrgb(iy+2, ix-1), pixrgb(iy+2, ix), pixrgb(iy+2, ix+1), pixrgb(iy+2, ix+2), x), y);
    }

    double bilinear(double vx, double vy) {
	if (vy < 1 || vy >= ih-1 || vx < 1 || vx >= iw-1) return 0;

	int x = (int)Math.floor(vx);
	int y = (int)Math.floor(vy);
	double u_ratio = vx - x, v_ratio = vy - y;
	double u_opposite = 1 - u_ratio, v_opposite = 1 - v_ratio;

	int p00 = (rgbBuf[(y + 0) * MAPSIZE + (x + 0)] >> 8) & 255;
	int p01 = (rgbBuf[(y + 0) * MAPSIZE + (x + 1)] >> 8) & 255;
	int p10 = (rgbBuf[(y + 1) * MAPSIZE + (x + 0)] >> 8) & 255;
	int p11 = (rgbBuf[(y + 1) * MAPSIZE + (x + 1)] >> 8) & 255;

	double p = (p00 * u_opposite + p01 * u_ratio) * v_opposite +
	    (p10 * u_opposite + p10 * u_ratio) * v_ratio;

	return p;
    }

    Vector3D[] refineSubpixelSub(Vector3D v0, Vector3D v1, int yNumStep, double xStepSize, int xNumStep, int sign, int thre, int range) {
	Vector3D ov, di, dj;

	Vector3D w = v1.minus(v0);
	w = (new Vector3D(-w.y, w.x, 0)).normalized();
	w = w.dot(sign);

	ov = v0.plus(w.dot(xStepSize * ((1 - xNumStep) / 2.0)));
	di = v1.minus(v0).dot(1.0 / (yNumStep+1.0));
	dj = w.dot(xStepSize);

	double sx2 = 0, sx = 0, syx = 0, sy = 0;

	for(int i=1;i<=yNumStep;i++) {
	    Vector3D v = ov.plus(di.dot(i));

	    //

	    double maxsum = -1e+100;
	    int maxindex = -1;
	    double sum = 0;

	    for(int j=0;j<xNumStep;j++) {
		double p = clamp(bicubic(v.x, v.y) - thre, -range, range);
		//double p = clamp(bilinear(v.x, v.y) - thre, -range, range);

		sum += p;
		if (sum > maxsum) {
		    maxsum = sum;
		    maxindex = j;
		}

		v = v.plus(dj);
	    }

	    double ax = i / (yNumStep+1.0);
	    double ay = xStepSize * (maxindex + 1 - xNumStep / 2.0);

	    sx  += ax;
	    sx2 += ax * ax;
	    syx += ay * ax;
	    sy  += ay;
	}

	double ca, cb;
	double d = 1.0 / (sx2 * yNumStep - sx * sx);

	ca = d * (yNumStep * syx - sx  * sy);
	cb = d * (     -sx * syx + sx2 * sy);

	Vector3D[] ret = { v0.plus(w.dot(cb)), v1.plus(w.dot(ca + cb)) };

	return ret;
    }

    void refineSubpixel(Vector3D[] refPoints, int index, int len, int sign, int black, int white) {
	Vector3D[][] a = new Vector3D[len][2];
	org.naokishibata.math.Line2D[] ln = new org.naokishibata.math.Line2D[len];

	for(int i=0;i < len;i++) {
	    Vector3D v0 = refPoints[index + ((i + 1) % len)];
	    Vector3D v1 = refPoints[index + i];
	    
	    a[i] = refineSubpixelSub(v0, v1, 8, 1.0/8, 32, sign, (int)((white + black) / 2), (int)((white - black) / 2));

	    ln[i] = new org.naokishibata.math.Line2D(new Vector2D(a[i][0].x, a[i][0].y), new Vector2D(a[i][1].x, a[i][1].y));
	}

	/*
	for(int i=0;i < len;i++) {
	    s0.drawLine(refPoints[index + ((i + 1) % len)].x, refPoints[index + ((i + 1) % len)].y, refPoints[index + i].x, refPoints[index + i].y, 2, Color.orange);
	}
	*/

	for(int i=0;i < len;i++) {
	    Vector2D w = org.naokishibata.math.Line2D.intersection(ln[i], ln[(i+1)%len]);
	    if (w != null) refPoints[index + ((i + 1) % len)] = new Vector3D(w.x + 0.5, w.y + 0.5, 0);
	}

	for(int i=0;i < len;i++) {
	    Color col = index == 4 ? Color.cyan : Color.red;
	    s0.drawLine(refPoints[index + ((i + 1) % len)].x, refPoints[index + ((i + 1) % len)].y, refPoints[index + i].x, refPoints[index + i].y, 0.5, col);
	}
    }

    public void doit(String fn, int iw0, int ih0, PrintWriter pw) throws Exception {
	CvCapture c0 = null;

	if (fn == null) {
	    c0 = CvCapture.cvCreateCameraCapture(0);
	    c0.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_WIDTH , iw0);
	    c0.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_HEIGHT, ih0);
	} else {
	    c0 = CvCapture.cvCreateFileCapture(fn);
	}

	iimg = c0.cvQueryFrame();
	iw = iimg.width();
	ih = iimg.height();

	imgBuf = new byte[iw * ih];

	s0 = new Scribble2D(iw, ih, true);
	s0.makeFrame("Camera 0");
	s0.setExitWhenClosed(true);
	s0.enableAntiAlias(true);

	CL.setExceptionsEnabled(true);
	cls = new CLHelper(0);

	/*
	cl_image_format iformat = new cl_image_format();
	iformat.image_channel_order = CL_R; iformat.image_channel_data_type = CL_FLOAT;
	clSrcImage = clCreateImage2D(cls.context, CL_MEM_READ_WRITE  | CL_MEM_ALLOC_HOST_PTR, new cl_image_format[] { iformat }, MAPSIZE, MAPSIZE, Sizeof.cl_float * MAPSIZE, null, null);
	*/

	cl_mem clSrcBuf = clCreateBuffer(cls.context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, iw * ih * 3, Pointer.to(iimg.getDirect()), null);
	cl_mem clBuf0 = clCreateBuffer(cls.context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, (1 << SHIFT) * (ih + 8) * Sizeof.cl_int, null, null);
	cl_mem clBuf1 = clCreateBuffer(cls.context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, (1 << SHIFT) * (ih + 8) * Sizeof.cl_int, null, null);
	cl_mem clBuf2 = clCreateBuffer(cls.context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, (1 << SHIFT) * (ih + 8) * Sizeof.cl_int, null, null);
	cl_mem clBuf3 = clCreateBuffer(cls.context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, (1 << SHIFT) * (ih + 8) * Sizeof.cl_int, null, null);
	cl_mem clDstBuf = clCreateBuffer(cls.context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, MAPSIZE * MAPSIZE * Sizeof.cl_int, null, null);

	cl_kernel kernel_clear0 = makeKernel("clear_C3.cl", "clear", iw, ih);
	clSetKernelArg(kernel_clear0, 0, Sizeof.cl_mem, Pointer.to(clBuf0));

	cl_kernel kernel_clear1 = makeKernel("clear_C3.cl", "clear", iw, ih);
	clSetKernelArg(kernel_clear1, 0, Sizeof.cl_mem, Pointer.to(clBuf1));

	cl_kernel kernel_clear2 = makeKernel("clear_C3.cl", "clear", iw, ih);
	clSetKernelArg(kernel_clear2, 0, Sizeof.cl_mem, Pointer.to(clBuf2));

	cl_kernel kernel_clear3 = makeKernel("clear_C3.cl", "clear", iw, ih);
	clSetKernelArg(kernel_clear3, 0, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass0 = makeKernel("pass0_C2.cl", "pass0", iw, ih);
	clSetKernelArg(kernel_pass0, 0, Sizeof.cl_mem, Pointer.to(clSrcBuf));
	clSetKernelArg(kernel_pass0, 1, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass0, 2, Sizeof.cl_mem, Pointer.to(clDstBuf));

	cl_kernel kernel_pass1 = makeKernel("pass1_C3.cl", "pass1", iw, ih);
	clSetKernelArg(kernel_pass1, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass1, 1, Sizeof.cl_mem, Pointer.to(clBuf1));
	clSetKernelArg(kernel_pass1, 2, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass1, 3, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass1a = makeKernel("pass1a_C6.cl", "pass1a", iw, ih);
	clSetKernelArg(kernel_pass1a, 0, Sizeof.cl_mem, Pointer.to(clBuf1));
	clSetKernelArg(kernel_pass1a, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass1a, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass1b = makeKernel("pass1b_Calib.cl", "pass1b", iw, ih);
	clSetKernelArg(kernel_pass1b, 0, Sizeof.cl_mem, Pointer.to(clBuf3));
	clSetKernelArg(kernel_pass1b, 1, Sizeof.cl_mem, Pointer.to(clBuf0));

	cl_kernel kernel_pass2 = makeKernel("pass2_C4.cl", "pass2", iw, ih);
	clSetKernelArg(kernel_pass2, 0, Sizeof.cl_mem, Pointer.to(clBuf0));

	cl_kernel kernel_pass3s = makeKernel("pass3s_C4.cl", "pass3s", iw, ih);
	clSetKernelArg(kernel_pass3s, 0, Sizeof.cl_mem, Pointer.to(clBuf0));

	cl_kernel kernel_pass4 = makeKernel("pass4_C6.cl", "pass4", iw, ih);
	clSetKernelArg(kernel_pass4, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass4, 1, Sizeof.cl_mem, Pointer.to(clBuf2));

	cl_kernel kernel_pass4a = makeKernel("pass4a_C6.cl", "pass4a", iw, ih);
	clSetKernelArg(kernel_pass4a, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass4a, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass4a, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass4be = makeKernel("pass4be_C6.cl", "pass4be", iw, ih);
	clSetKernelArg(kernel_pass4be, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass4be, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass4be, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass4d = makeKernel("pass4d_C6.cl", "pass4d", iw, ih);
	clSetKernelArg(kernel_pass4d, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass4d, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass4d, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass5 = makeKernel("pass5_C6.cl", "pass5", iw, ih);
	clSetKernelArg(kernel_pass5, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass5, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass5, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass6 = makeKernel("pass6_C6.cl", "pass6", iw, ih);
	clSetKernelArg(kernel_pass6, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass6, 1, Sizeof.cl_mem, Pointer.to(clBuf2));
	clSetKernelArg(kernel_pass6, 2, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass7 = makeKernel("pass7_C6.cl", "pass7", iw, ih);
	clSetKernelArg(kernel_pass7, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	clSetKernelArg(kernel_pass7, 1, Sizeof.cl_mem, Pointer.to(clBuf3));

	cl_kernel kernel_pass9 = null;
	if (enable_pass9) {
	    kernel_pass9 = makeKernel("pass9_C2.cl", "pass9", iw, ih);
	    clSetKernelArg(kernel_pass9, 0, Sizeof.cl_mem, Pointer.to(clBuf0));
	    clSetKernelArg(kernel_pass9, 1, Sizeof.cl_mem, Pointer.to(clDstBuf));
	}

	//

	rgbBuf = new int[MAPSIZE * MAPSIZE];
	int[] coordBuf = new int[MAPSIZE * ih];
	BufferedImage bimg = new BufferedImage(iw, ih, BufferedImage.TYPE_INT_ARGB);

	long lastTime = System.currentTimeMillis();
	int frameCount = 0;

	cls.showInfo();

	Vector3D[][] refPoints = new Vector3D[20][20];
	int refCount = 0;
	Vector3D[] sumRefPoints = new Vector3D[20];

	for(int i=0;i<20;i++) {
	    for(int j=0;j<20;j++) refPoints[i][j] = new Vector3D(0, 0, 0);
	}

	System.out.println("You need to first print out pat/calib.pdf.");
	System.out.println("Move the focus to the capture window, and head the camera towards the printed image.");
	System.out.println("Tips : Try capturing image at an angle. Auto-focus may affect the camera parameters.");

	System.out.println();
	System.out.println("Press ENTER to capture an image");
	System.out.println("Press ESC to quit");

	for(;;) {
	    if (!benchMode) {
		iimg = c0.cvRetrieveFrame();
		c0.cvGrabFrame();
	    }

	    if (iimg == null) break;

   	    clEnqueueWriteBuffer(cls.commandQueue, clSrcBuf, CL_FALSE, 0, iw * ih * 3, Pointer.to(iimg.getDirect()), 0, null, null);

	    // clear all buffers

	    final int wsx = -1, wsy = -1;

	    cls.run2D(kernel_clear0, MAPSIZE, ih+8, wsx, wsy);
	    cls.run2D(kernel_clear1, MAPSIZE, ih+8, wsx, wsy);
	    cls.run2D(kernel_clear2, MAPSIZE, ih+8, wsx, wsy);
	    cls.run2D(kernel_clear3, MAPSIZE, ih+8, wsx, wsy);

	    // pass 0

	    cls.run2D(kernel_pass0, iw, ih, wsx, wsy);

	    if (!benchMode) {
		clEnqueueReadBuffer(cls.commandQueue, clDstBuf, CL_TRUE, 0, MAPSIZE * MAPSIZE * Sizeof.cl_int, Pointer.to(rgbBuf), 0, null, null);
	    }

	    // Edge detection / thinning

	    cls.run2D(kernel_pass1 , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass1a, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass1b, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);

	    // CCL

	    cls.run2D(kernel_pass2 , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass3s, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);

	    if (enable_pass9) {
		cls.run2D(kernel_pass9, iw, ih, wsx, wsy);
		clEnqueueReadBuffer(cls.commandQueue, clDstBuf, CL_TRUE, 0, MAPSIZE * MAPSIZE * Sizeof.cl_int, Pointer.to(rgbBuf), 0, null, null);
	    }

	    cls.run2D(kernel_clear1, MAPSIZE, ih, wsx, wsy);
	    cls.run2D(kernel_clear2, MAPSIZE, ih, wsx, wsy);
	    cls.run2D(kernel_clear3, MAPSIZE, ih, wsx, wsy);

	    // Discard small blobs

	    cls.run2D(kernel_pass4  , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass4a , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass4be, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_clear3 , MAPSIZE, ih, wsx, wsy);
	    cls.run2D(kernel_pass4d , EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass4be, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_clear3 , MAPSIZE, ih, wsx, wsy);
	    cls.run2D(kernel_clear2 , MAPSIZE, ih, wsx, wsy);

	    // Obtain 8 points for each blob

	    cls.run2D(kernel_pass4, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass5, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);
	    cls.run2D(kernel_pass6, EDGE, EDGE, iw - EDGE*2, ih - EDGE*2, wsx, wsy);

	    cls.run1D(kernel_pass7, 256 * 8, wsx);

	    clEnqueueReadBuffer(cls.commandQueue, clBuf3, CL_TRUE, 0, MAPSIZE * ih * Sizeof.cl_int, Pointer.to(coordBuf), 0, null, null);

	    cls.flush();

	    if (!benchMode) {
		bimg.setRGB(0, 0, iw, ih, rgbBuf, 0, MAPSIZE);

		s0.setFreezed(true);
		s0.drawImage(0, 0, bimg);

		//

		LinkedHashSet<Node> allNodes = new LinkedHashSet<Node>();

		//

		double ox=-1, oy=-1, xx=0, xy=0, yx=0, yy=0;
		double xx2=0, xy2=0, yx2=0, yy2=0;
		int nInput = 0, nOutput = 0;

		double box=-1, boy=-1, bxx=0, bxy=0, byx=0, byy=0;
		double bxx2=0, bxy2=0, byx2=0, byy2=0;
		int bsize = -1;
		Vector3D bvo = null, bvx = null, bvy = null;

		for(int p=0;p<MAPSIZE*ih/16;p++) {
		    if (coordBuf[p*16] == 0 || skip_cpu_part) break;

		    Vector2D[] vt = new Vector2D[16];
		    int[] vatr = new int[16];

		    nInput++;

		    int xmin = MAPSIZE, xmax = -1, ymin = MAPSIZE, ymax = -1;
		    for(int i=0;i<8;i++) {
			int x0 = coordBuf[p*16+8+i] & ((1 << SHIFT)-1), y0 = (coordBuf[p*16+8+i] >> SHIFT) & ((1 << SHIFT)-1);
			vt[i] = vt[i+8] = new Vector2D(x0, y0);
			vatr[i] = vatr[i+8] = coordBuf[p*16+i];
			xmin = min(xmin, x0); xmax = max(xmax, x0);
			ymin = min(ymin, y0); ymax = max(ymax, y0);
		    }

		    /*
		    for(int i=0;i<8;i++) {
			Color c = Color.white;
			s0.drawLine(vt[i].x, vt[i].y, vt[i+1].x, vt[i+1].y, 1, c);
		    }
		    */

		    int size = max(xmax-xmin, ymax-ymin);

		    Node n = new Node(p, vt, new Vector2D(xmin, ymin), new Vector2D(xmax, ymax));

		    allNodes.add(n);
		}

		for(Node n : allNodes) {
		    for(Node m : allNodes) {
			if (n.ul.x < m.ul.x && n.ul.y < m.ul.y && n.lr.x > m.lr.x && n.lr.y > m.lr.y) {
			    final double d = 1.8;
			    if (n.ar * (1.0 / d) < m.ar && m.ar < n.ar * d) {
				n.children.add(m);
			    }
			}
		    }
		}

		Node frameNode = null;

		for(Node n : allNodes) {
		    //System.out.println(n + " : " + n.children.size());

		    if (n.children.size() != 4) continue;

		    boolean b = true;

		    for(Node m : n.children) {
			if (m.children.size() != 0) { b = false; break; }
		    }

		    if (!b) continue;

		    frameNode = n;
		}

		if (frameNode != null) {
		    int[] nn = {-1, -1, -1, -1};

		    {
			double dmax = 0;
			for(int i=0;i<8;i++) {
			    for(int j=0;j<8;j++) {
				double d = Vector2D.distance(frameNode.vt[i], frameNode.vt[j]);
				if (d > dmax) {
				    dmax = d;
				    nn[0] = i; nn[1] = j;
				}
			    }
			}
		    }

		    //drawX(s0, frameNode.vt[nn[0]], Color.green);
		    //drawX(s0, frameNode.vt[nn[1]], Color.green);

		    Vector2D fcenter = Vector2D.midpoint(frameNode.vt[nn[0]], frameNode.vt[nn[1]]);
		    
		    Vector2D north = new Vector2D(frameNode.vt[nn[0]].x - frameNode.vt[nn[1]].x, frameNode.vt[nn[0]].y - frameNode.vt[nn[1]].y);
		    north = north.normalize();

		    Vector2D east = new Vector2D(frameNode.vt[nn[0]].y - frameNode.vt[nn[1]].y, frameNode.vt[nn[1]].x - frameNode.vt[nn[0]].x);
		    east = east.normalize();

		    Node triNode = null;

		    {
			double rmin = 1e+100;

			for(Node n : frameNode.children) {
			    double dmin = 1e+100, dmax = -1e+100;

			    for(int i=0;i<8;i++) {
				double d = Vector2D.distance(n.vt[i], fcenter);
				dmin = Math.min(d, dmin);
				dmax = Math.max(d, dmax);
			    }

			    double r = dmax / dmin;

			    if (r < rmin) {
				triNode = n;
				rmin = r;
			    }
			}
		    }

		    /*
		    for(int i=0;i<8;i++) {
			Color c = Color.red;
			s0.drawLine(frameNode.vt[i].x, frameNode.vt[i].y, frameNode.vt[i+1].x, frameNode.vt[i+1].y, 2, c);
		    }

		    for(Node n : frameNode.children) {
			Color c = (n == triNode) ? Color.cyan : Color.blue;

			for(int i=0;i<8;i++) {
			    s0.drawLine(n.vt[i].x, n.vt[i].y, n.vt[i+1].x, n.vt[i+1].y, 2, c);
			}
		    }
		    */

		    {
			double dmax2 = -1e+100, dmax3 = -1e+100;
			for(int i=0;i<8;i++) {
			    Vector2D w;

			    w = frameNode.vt[i].minus(fcenter);
			    double c = east.dot(w);

			    if (c > dmax2) {
				dmax2 = c;
				nn[2] = i;
			    }

			    if (-c > dmax3) {
				dmax3 = -c;
				nn[3] = i;
			    }
			}
		    }

		    int[] nn2 = {-1, -1, -1, -1};

		    {
			int vn = -1, vs = -1;
			double dmin = 1e+100, dmax = -1e+100;
			Vector2D tc = Vector2D.midpoint(triNode.ul, triNode.lr);
			for(int i=0;i<4;i++) {
			    double d = Vector2D.distance(frameNode.vt[nn[i]], tc);
			    if (d < dmin) {
				dmin = d;
				vn = nn[i];
			    }
			    if (d > dmax) {
				dmax = d;
				vs = nn[i];
			    }
			}

			north = new Vector2D(frameNode.vt[vn].x - frameNode.vt[vs].x, frameNode.vt[vn].y - frameNode.vt[vs].y);
			north = north.normalize();

			east = new Vector2D(frameNode.vt[vn].y - frameNode.vt[vs].y, frameNode.vt[vs].x - frameNode.vt[vn].x);
			east = east.normalize();

			nn2[0] = vn; nn2[2] = vs;

			//

			double dmax2 = -1e+100, dmax3 = -1e+100;
			for(int i=0;i<8;i++) {
			    Vector2D w;

			    w = frameNode.vt[i].minus(fcenter);
			    double c = east.dot(w);

			    if (c > dmax2) {
				dmax2 = c;
				nn2[3] = i;
			    }

			    if (-c > dmax3) {
				dmax3 = -c;
				nn2[1] = i;
			    }
			}

			east = new Vector2D(frameNode.vt[nn2[1]].x - frameNode.vt[nn2[3]].x, frameNode.vt[nn2[1]].y - frameNode.vt[nn2[3]].y);
			east = east.normalize();
		    }

		    s0.drawLine(fcenter.x, fcenter.y, fcenter.x + north.x * 20, fcenter.y + north.y * 20, 3, Color.blue);
		    s0.drawLine(fcenter.x, fcenter.y, fcenter.x + east.x * 20, fcenter.y + east.y * 20, 3, Color.red);

		    /*
		    s0.drawString("N", frameNode.vt[nn2[0]].x, frameNode.vt[nn2[0]].y, Scribble2D.CENTER, 25, Color.green);
		    s0.drawString("E", frameNode.vt[nn2[1]].x, frameNode.vt[nn2[1]].y, Scribble2D.CENTER, 25, Color.green);
		    s0.drawString("S", frameNode.vt[nn2[2]].x, frameNode.vt[nn2[2]].y, Scribble2D.CENTER, 25, Color.green);
		    s0.drawString("W", frameNode.vt[nn2[3]].x, frameNode.vt[nn2[3]].y, Scribble2D.CENTER, 25, Color.green);
		    */

		    //drawX(s0, frameNode.vt[nn[2]], Color.green);
		    //s0.drawString("f2", frameNode.vt[nn[2]].x, frameNode.vt[nn[2]].y, Scribble2D.CENTER, 15, Color.white);
		    //drawX(s0, frameNode.vt[nn[3]], Color.green);
		    //s0.drawString("f3", frameNode.vt[nn[3]].x, frameNode.vt[nn[3]].y, Scribble2D.CENTER, 15, Color.white);

		    for(int i=0;i<4;i++) {
			refPoints[refCount][i] = new Vector3D(frameNode.vt[nn2[i]].x, frameNode.vt[nn2[i]].y, 0);
		    }

		    frameNode.nn = nn;
		    frameNode.cn = new Node[4];

		    {
			double dmax0 = -1e+100, dmax1 = -1e+100, dmax2 = -1e+100, dmax3 = -1e+100;
			for(Node n : frameNode.children) {
			    Vector2D w;

			    Vector2D ncenter = Vector2D.midpoint(n.ul, n.lr);

			    //drawX(s0, ncenter, Color.pink);

			    w = ncenter.minus(fcenter);
			    double c = north.dot(w);

			    if (c > dmax0) {
				dmax0 = c;
				frameNode.cn[0] = n;
			    }

			    if (-c > dmax1) {
				dmax1 = -c;
				frameNode.cn[2] = n;
			    }

			    c = east.dot(w);

			    if (c > dmax2) {
				dmax2 = c;
				frameNode.cn[1] = n;
			    }

			    if (-c > dmax3) {
				dmax3 = -c;
				frameNode.cn[3] = n;
			    }
			}

		    }

		    for(int j=0;j<4;j++) {
			Node n = frameNode.cn[j];
			Vector2D center = Vector2D.midpoint(n.ul, n.lr);

			double dmax0 = -1e+100, dmax1 = -1e+100, dmax2 = -1e+100, dmax3 = -1e+100;
			for(int i=0;i<8;i++) {
			    Vector2D w;

			    w = n.vt[i].minus(center);
			    double c = north.dot(w);

			    if (c > dmax0) {
				dmax0 = c;
				nn[0] = i;
			    }

			    if (-c > dmax1) {
				dmax1 = -c;
				nn[2] = i;
			    }

			    c = east.dot(w);

			    if (c > dmax2) {
				dmax2 = c;
				nn[1] = i;
			    }

			    if (-c > dmax3) {
				dmax3 = -c;
				nn[3] = i;
			    }
			}

			n.nn = nn;

			//drawX(s0, n.vt[nn[0]], Color.green);
			//drawX(s0, n.vt[nn[1]], Color.green);
			//drawX(s0, n.vt[nn[2]], Color.green);
			//drawX(s0, n.vt[nn[3]], Color.green);

			for(int k=0;k<4;k++) {
			    refPoints[refCount][4+j*4+k] = new Vector3D(n.vt[nn[k]].x, n.vt[nn[k]].y, 0);
			}
		    }

		    {
			Vector3D w;

			w = Vector3D.midpoint(refPoints[refCount][0], refPoints[refCount][4]);
			int black = (rgbBuf[((int)w.y) * MAPSIZE + ((int)w.x)] >> 8) & 255;

			w = Vector3D.midpoint(refPoints[refCount][8], refPoints[refCount][10]);
			int white = (rgbBuf[((int)w.y) * MAPSIZE + ((int)w.x)] >> 8) & 255;

			refineSubpixel(refPoints[refCount],  0, 4, -1, black, white);

			refPoints[refCount][6] = refPoints[refCount][7];
			refineSubpixel(refPoints[refCount],  4, 3,  1, black, white);
			refPoints[refCount][7] = refPoints[refCount][6];

			refineSubpixel(refPoints[refCount],  8, 4,  1, black, white);
			refineSubpixel(refPoints[refCount], 12, 4,  1, black, white);
			refineSubpixel(refPoints[refCount], 16, 4,  1, black, white);
		    }

		    /*
		    for(int i=0;i<4;i++) {
			drawX(s0, refPoints[refCount][i], Color.blue);
		    }
		    */

		    for(int i=0;i<20;i++) {
			sumRefPoints[i] = new Vector3D(0, 0, 0);
			for(int j=0;j<20;j++) {
			    sumRefPoints[i] = sumRefPoints[i].plus(refPoints[j][i]);
			}
			sumRefPoints[i] = sumRefPoints[i].dot(1.0/20);
		    }

		    for(int i=0;i<20;i++) {
			if (i == 6) continue;
			s0.drawString("" + i, sumRefPoints[i].x, sumRefPoints[i].y, Scribble2D.CENTER, 20, Color.white);
		    }

		    if ((s0.keyState() & (KeySensor.ENTER)) != 0) {
			pw.println("BEGIN " + iw + "," + ih);
			for(int i=0;i<20;i++) {
			    pw.println("" + sumRefPoints[i].x + "," + sumRefPoints[i].y);
			}
			pw.println("END");

			while((s0.keyState() & (KeySensor.ENTER)) != 0) {
			    try { Thread.sleep(100); } catch(Exception ex) {}
			}

			System.out.println("Image captured");
			System.out.println("Press ESC to quit");
		    }

		    refCount = (refCount + 1) % 20;
		}

		//System.out.println("nCoord:" + nInput);
		s0.setFreezed(false);
	    }

	    long time = System.currentTimeMillis();
	    frameCount++;

	    if (time - lastTime > 1000) {
		lastTime = lastTime + 1000;
		frameCount = 0;
	    }

	    if ((s0.keyState() & (KeySensor.ESC)) != 0) {
		pw.close();
		break;
	    }
	}
    }

    static int max(int a, int b) { return a > b ? a : b; }
    static int min(int a, int b) { return a < b ? a : b; }

    static double clamp(double x, double min, double max) {
	if (x < min) return min;
	if (x > max) return max;
	return x;
    }

    //

    static class EularAngles {
	public final double phi, theta, psi;
	Matrix mat = null, imat = null;

	public EularAngles(double phi, double theta, double psi) {
	    this.phi = phi;
	    this.theta = theta;
	    this.psi = psi;
	}

	public String toString() {
	    String s;
	    s = "[EularAngles " + phi + ", " + theta + ", " + psi + "]";
	    return s;
	}

	void makeMat() {
	    FastMath.double2 d2;

	    d2 = FastMath.sincos(psi);
	    double[][] a1 = {
		{ d2.y,  0, d2.x, 0},
		{    0,  1,    0, 0},
		{-d2.x,  0, d2.y, 0},
		{    0,  0,    0, 1},
	    };

	    d2 = FastMath.sincos(theta);
	    double[][] a2 = {
		{ 1,    0,    0, 0},
		{ 0, d2.y,-d2.x, 0},
		{ 0, d2.x, d2.y, 0},
		{ 0,    0,    0, 1},
	    };
	
	    d2 = FastMath.sincos(phi);
	    double[][] a3 = {
		{ d2.y,-d2.x, 0, 0},
		{ d2.x, d2.y, 0, 0},
		{    0,    0, 1, 0},
		{    0,    0, 0, 1},
	    };

	    mat = (new Matrix(a3)).times((new Matrix(a2)).times(new Matrix(a1)));
	}

	Matrix mat() {
	    if (mat == null) makeMat();

	    return new Matrix(mat);
	}

	Matrix imat() {
	    if (imat == null) {
		if (mat == null) makeMat();
		imat = mat.inversion();
	    }

	    return new Matrix(imat);
	}
    }

    static class CamParam {
	final Vector3D pos;    // camera position
	final EularAngles ea;  // camera orientation

	Matrix mat = null, imat = null;

	public String toString() {
	    String s;
	    s = "[CamParam " + pos + ", " + ea + "]";
	    return s;
	}

	CamParam(Vector3D pos, double phi, double theta, double psi) {
	    this.pos = pos;
	    this.ea = new EularAngles(phi, theta, psi);
	}

	Matrix mat() {
	    if (mat == null) {
		double[][] a = {
		    {  1,  0,  0, pos.x},
		    {  0,  1,  0, pos.y},
		    {  0,  0,  1, pos.z},
		    {  0,  0,  0,     1},
		};

		mat = (new Matrix(a)).times(ea.mat());
	    }

	    return mat;
	}

	Matrix imat() {
	    if (imat == null) imat = mat().inversion();

	    return imat;
	}

	Vector3D dirToCam(Vector3D dir) {
	    Matrix cam = mat().times(new VectorN(dir.x, dir.y, dir.z, 1));
	    return new Vector3D(cam.v(0, 0), cam.v(0, 1), cam.v(0, 2));
	}

	Vector3D camToDir(Vector3D cam) {
	    Matrix dir = imat().times(new VectorN(cam.x, cam.y, cam.z, 1));
	    return new Vector3D(dir.v(0, 0), dir.v(0, 1), dir.v(0, 2));
	}
    }

    static double[][] refpat = {
	{    - 2.5,    - 2.5}, {      2.5,    - 2.5}, {      2.5,      2.5}, {    - 2.5,      2.5},
	{ -1 - 0.5, -1 - 0.5}, { -1 + 0.5, -1 - 0.5}, { -1 + 0.5, -1 + 0.5}, { -1 - 0.5, -1 + 0.5},
	{  1 - 0.5, -1 - 0.5}, {  1 + 0.5, -1 - 0.5}, {  1 + 0.5, -1 + 0.5}, {  1 - 0.5, -1 + 0.5},
	{  1 - 0.5,  1 - 0.5}, {  1 + 0.5,  1 - 0.5}, {  1 + 0.5,  1 + 0.5}, {  1 - 0.5,  1 + 0.5},
	{ -1 - 0.5,  1 - 0.5}, { -1 + 0.5,  1 - 0.5}, { -1 + 0.5,  1 + 0.5}, { -1 - 0.5,  1 + 0.5},
    };

    static class ParamFinder implements ConjGrad.Instance {
	final Vector3D[][] refPoints;
	final int iw, ih, mode;
	final double paspect;
	int count = 0;

	ParamFinder(int mode, Vector3D[][] refPoints, int iw, int ih, double paspect) {
	    this.refPoints = refPoints;
	    this.iw = iw;
	    this.ih = ih;
	    this.mode = mode;
	    this.paspect = paspect;
	}

	Vector3D lenseWarp(Vector3D v, double distance, double aspect, double distortion) {
	    Vector3D w = new Vector3D(v.x * distance / v.z, v.y * distance / v.z, distance);
	    {
		double vx = w.x - iw/2;
		double vy = w.y - ih/2;
		double d = vx * vx + vy * vy;
		double m = d / (ih * ih) * distortion;
		vx *= 1 + m;
		vy *= 1 + m;
		w = new Vector3D(iw/2 + vx, ih/2 + vy, w.z);
	    }
	    w = new Vector3D(w.x, (w.y - ih/2) * aspect + ih/2, w.z);
	    return w;
	}

	public double value(VectorN v) {
	    double ret = 0;

	    double aov = mode == 0 ? v.v(v.m()-1) : v.v(v.m()-3);
	    double aspect = v.v(v.m()-2);
	    double distortion = v.v(v.m()-1);
	    double distance = iw / (2 * Math.tan(aov/2));

	    for(int j=0;j<refPoints.length;j++) {
		CamParam cp = new CamParam(new Vector3D(v.v(j*6+0)*100.0, v.v(j*6+1)*100.0, v.v(j*6+2)*100.0), v.v(j*6+3), v.v(j*6+4), v.v(j*6+5));
		double sign = 1; //refPoints[j][0].x > refPoints[j][1].x ? -1 : 1;

		for(int i=0;i<20;i++) {
		    if (i == 6) continue;
		    Vector3D w = cp.dirToCam(new Vector3D(paspect * sign * refpat[i][0] * iw / 3, refpat[i][1] * iw / 3, 0));

		    switch(mode) {
		    case 2: w = lenseWarp(w, distance, aspect, distortion); break;
		    case 1: w = lenseWarp(w, distance, aspect, 0); break;
		    default: w = lenseWarp(w, distance, 1, 0); break;
		    } 

		    //System.out.println(refPoints[j][i] + ", (" + refpat[i][0] + "," + refpat[i][1] + ")");
		    
		    double e = (w.x - (refPoints[j][i].x - iw/2)) * (w.x - (refPoints[j][i].x - iw/2)) + (w.y - (refPoints[j][i].y - ih/2)) * (w.y - (refPoints[j][i].y - ih/2));
		    ret += e * e;
		}
	    }

	    //System.exit(0);


	    if ((count++) % 10000 == 0) {
		//System.out.println("v = " + v);
		//System.out.println("aov = " + aov);
		//System.out.println("distance = " + distance);
		System.out.println(ret);
	    }

	    return ret;
	}

	public VectorN gradient(VectorN v) {
	    return MathUtils.gradientAprx(this, v, 0.001);
	}

	public VectorN gradient2(VectorN v) {
	    return MathUtils.gradient2Aprx(this, v, 0.001);
	}

	public boolean finish(VectorN v) {
	    return gradient(v).length() < 0.00001;
	}
    }

    public static void main(String[] args) throws Exception {
	HashMap<String, Integer> options = new HashMap<String, Integer>();
	options.put("--capture", 1);
	options.put("--videofile", 1);
	options.put("--size", 2);
	options.put("--findparam", 2);
	options.put("--drawpat", 0);

	HashMap<String, String[]> cl = MiscUtils.analyseCommandLine(args, options);

	if (cl == null) {
	    System.out.println("This is a program for camera calibration. " +
			       "You need to execute this program at least twice. " +
			       "First, you need to capture the image. " +
			       "Then, you need to find the camera parameter.");
	    System.out.println();
	    System.out.println("To capture an image, you need to first print out pat/calib.pdf.");
	    System.out.println("And then use --capture option as shown below");
	    System.out.println();
	    System.out.println("Step 1 : java Calib --capture <param file name>");
	    System.out.println("  Capture the calibration pattern from camera");
	    System.out.println("  --size <width> <height>         Specify image size");
	    System.out.println("  --videofile <video file name>   Capture video from file");
	    System.out.println("  --drawpat                       Draw calibration pattern");
	    System.out.println();
	    System.out.println("To find the camera parameter, use --findparam option as shown below");
	    System.out.println();
	    System.out.println("Step 2 : java Calib --findparam <param file name> <printer aspect ratio>");
	    System.out.println("  Find the lense parameters");
	    System.out.println("  If you don't know the printer aspect ratio, try 1.00");
	    System.exit(-1);
	}

	String videofn = null, paramfn = null;
	int width = 1024, height = 768;
	boolean modeCapture = false, modeFind = false, modeDraw = false;
	double paspect = 1.0;

	for(String k : cl.keySet()) {
	    String[] a = cl.get(k);
	    if (k.equals("--capture")) {
		modeCapture = true;
		paramfn = a[0];
	    } else if (k.equals("--findparam")) {
		modeFind = true;
		paramfn = a[0];
		paspect = Double.parseDouble(a[1]);
	    } else if (k.equals("--videofile")) {
		videofn = a[0];
	    } else if (k.equals("--size")) {
		width = Integer.parseInt(a[0]);
		height = Integer.parseInt(a[1]);
	    } else if (k.equals("--drawpat")) {
		modeDraw = true;
	    }
	}

	if (modeDraw) {
	    int bs = 128;

	    Scribble2D s = new Scribble2D(bs*5, bs*5);

	    s.makeFrame();

	    int xoffset = 0;
	    int yoffset = 0;

	    int c = 0;
	    Color col = new Color(c, c, c);

	    for(int y=0;y<5;y++) {
		for(int x=0;x<5;x++) {
		    boolean b = false;

		    if (y == 0) b = true;
		    if (y == 4) b = true;
		    if (x == 0) b = true;
		    if (x == 4) b = true;

		    if ((x & 1) == 0) b = true;
		    if ((y & 1) == 0) b = true;

		    if (b) {
			s.fillBox(xoffset + bs*x, yoffset + bs*y, xoffset + bs*(x+1), yoffset + bs*(y+1), col);
		    }
		}
	    }

	    s.fillLines(col, bs*2, bs*1, bs*1, bs*2, bs*2, bs*2);

	    s.saveImage("png", new File("calib.png"));

	    System.exit(0);
	}

	if (modeCapture == modeFind) {
	    System.err.println("Specify only one of --capture or --findparam");
	    System.exit(-1);
	}

	if (modeCapture) {
	    Calib instance = new Calib();

	    instance.doit(videofn, width, height, new PrintWriter(new File(paramfn)));

	    System.exit(0);
	}


	LineNumberReader lnr = new LineNumberReader(new FileReader(paramfn));
	LinkedHashSet<Vector3D[]> set = new LinkedHashSet<Vector3D[]>();
	int iw = 0, ih = 0;
	double aov = 55.0 / 180.0 * Math.PI;

	for(;;) {
	    String s;

	    s = lnr.readLine();
	    if (s == null) break;
	    if (!s.startsWith("BEGIN ")) break;

	    {
		String[] ss = s.substring(6).split(",");
	        int iw0 = Integer.parseInt(ss[0]);
		int ih0 = Integer.parseInt(ss[1]);

		if (iw == 0) {
		    iw = iw0; ih = ih0;
		} else {
		    if (iw != iw0 || ih != ih0) {
			System.err.println("resolution mismatch");
			System.exit(-1);
		    }
		}
	    }

	    Vector3D[] a = new Vector3D[20];
	    for(int i=0;i<20;i++) {
		s = lnr.readLine();
		String[] ss = s.split(",");
		a[i] = new Vector3D(Double.parseDouble(ss[0]), Double.parseDouble(ss[1]), 0);
	    }
	    set.add(a);

	    s = lnr.readLine();
	    if (!s.startsWith("END")) break;
	}

	Vector3D[][] refPointArray = set.toArray(new Vector3D[0][0]);
	double[] an = new double[6*refPointArray.length + 3];

	for(int j=0;j<refPointArray.length;j++) {
	    Vector3D[][] rpa2 = { refPointArray[j] };
	    ParamFinder pf = new ParamFinder(0, rpa2, iw, ih, paspect);
	    VectorN x0 = ConjGrad.execute(pf, new VectorN(0, 0, iw / (2 * Math.atan(aov/2)) / 100.0, 0, 0, 0, aov), 0.0001, 0.00001, 10000);
	    for(int i=0;i<6;i++) an[j*6+i] = x0.v(i);
	    aov = x0.v(6);
	}

	an[an.length - 3] = aov;
	an[an.length - 2] = 1.0;
	an[an.length - 1] = 0.0;

	VectorN vn = new VectorN(an);

	ParamFinder pf = new ParamFinder(2, refPointArray, iw, ih, paspect);
	vn = ConjGrad.execute(pf, vn, 0.0001, 0.00001, 5000);

	System.out.println();
	System.out.println("*** Evaluation = " + pf.value(vn));
	System.out.println("*** Distortion = " + vn.v(an.length - 1));
	System.out.println("*** Aspect     = " + vn.v(an.length - 2));
	System.out.println("*** AOV        = " + (vn.v(an.length - 3)/Math.PI*180));

	System.exit(0);
    }
}
