package org.naokishibata.generic;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.regex.*;

public class MiscUtils {
    public static Object mutexDirectBuffer = new Object();

    public static <T> T getFirst(Iterable<T> c) {
	T ret = null;

	for(T o : c) {
	    ret = o;
	    break;
	}

	return ret;
    }

    public static <T> boolean isEmpty(Iterable<T> c) {
	boolean ret = true;

	for(T o : c) {
	    ret = false;
	    break;
	}

	return ret;
    }

    public static <T, U extends T> void addAll(Collection<T> c, Iterable<U> i) {
	for(U e : i) c.add(e);
    }

    public static <T, U extends T> void addAll(Collection<T> c, U ... e) {
	for(int i=0;i<e.length;i++) c.add(e[i]);
    }

    public static <T> T[] toArray(Iterable<T> it, T[] a) {
	LinkedList<T> list = new LinkedList<T>();
	addAll(list, it);
	return list.toArray(a);
    }

    public static class WrappedIterable<T> implements Iterable<T> {
	private Iterable<T> it;

	public WrappedIterable(Iterable<T> it0) {
	    it = it0;
	}

	public Iterator<T> iterator() {
	    return it.iterator();
	}

	public String toString() {
	    String ret = "[";
	    for(T t : this) ret = ret + t + " ";
	    return ret + "]";
	}
    }

    public static <T> Iterable<T> iterableWrapper(Iterable<T> it) {
	return new WrappedIterable<T>(it);
    }

    public static class IteratorConverter<T> implements Iterable<T> {
	private Iterator<T> it;

	public IteratorConverter(Iterator<T> it) { this.it = it; }

	public Iterator<T> iterator() { return it; }

	public String toString() {
	    String ret = "[";
	    for(T t : this) ret = ret + t + " ";
	    return ret + "]";
	}
    }

    public static <T> Iterable<T> iteratorToIterable(Iterator<T> it) { return new IteratorConverter<T>(it); }

    public static String replaceString(String str, String pattern, String replace) {
	int len = pattern.length();
	String sb = "";
	int found = -1;
	int start = 0;

	while( (found = str.indexOf(pattern, start) ) != -1) {
	    sb = sb + str.substring(start, found);
	    sb = sb + replace;
	    start = found + len;
	}

	sb = sb + str.substring(start);

	return sb;
    }

    public static String substringFromEnd(String str, int index) {
	if (index > str.length()) index = str.length();
	return str.substring(str.length() - index, str.length());
    }

    public static String sprintf(String format, Object... args) {
	StringWriter sr = new StringWriter();
	PrintWriter pw = new PrintWriter(sr);

	pw.printf(format, args);

	return sr.toString();
    }

    public static String getThreadStarterClassName() {
	StackTraceElement[] elm = Thread.currentThread().getStackTrace();

	return elm[elm.length-1].getClassName();
    }

    static Thread getMainThreadRecurse(ThreadGroup g) {
	if (g.getName().equals("main")) {
	    int ac = g.activeCount()*2;
	    Thread[] a = new Thread[ac];

	    g.enumerate(a);

	    for(int i=0;i<a.length;i++) {
		if (a[i] != null) {
		    if (a[i].getName().equals("main")) return a[i];
		}
	    }
	}

	{
	    int ac = g.activeGroupCount()*2;
	    ThreadGroup[] a = new ThreadGroup[ac];

	    g.enumerate(a);

	    for(int i=0;i<a.length;i++) {
		if (a[i] != null) {
		    Thread th = getMainThreadRecurse(a[i]);
		    if (th != null) return th;
		}
	    }
	}

	return null;
    }

    public static Thread getMainThread() {
	ThreadGroup root = Thread.currentThread().getThreadGroup().getParent();
	while (root.getParent() != null) root = root.getParent();

	return getMainThreadRecurse(root);
    }

    public static String getMainClassName() {
	Thread th = getMainThread();

	if (th == null) return "Unknown";

	StackTraceElement[] se = th.getStackTrace();

	return se[se.length-1].getClassName();
    }

    public static String getCallerClassName() {
	StackTraceElement[] se = Thread.currentThread().getStackTrace();

	if (se.length >= 4) return se[3].getClassName();

	return "Unknown";
    }

    public static String readWholeFileAsAString(File f) throws FileNotFoundException, IOException {
	String ret = "";

	BufferedReader br = new BufferedReader(new FileReader(f));
	for(;;) {
	    String s = br.readLine();
	    if (s == null) break;
	    ret = ret + s + "\n";
	}

	return ret;
    }

    public static String readWholeFileAsAString(String fn) throws FileNotFoundException, IOException {
	String ret = "";

	BufferedReader br = new BufferedReader(new InputStreamReader(ClassLoader.getSystemClassLoader().getResourceAsStream(fn)));
	for(;;) {
	    String s = br.readLine();
	    if (s == null) break;
	    ret = ret + s + "\n";
	}

	return ret;
    }

    public static final String regexFPNumber = "[-+]?[0-9]*\\.?(?:(?:[0-9]*)|(?:[0-9]+(?:(?:(?:[eE][-+]?)|(?:[eE][-+]?[0-9]+)))?))";

    static LRUMap<String, Pattern> regexMap = new LRUMap<String, Pattern>(100);

    public static String[] regexCapture(String input, String regex) {
	Matcher m = null;

	synchronized(regexMap) {
	    Pattern p = regexMap.get(regex);
	    if (p == null) {
		p = Pattern.compile(regex);
		regexMap.put(regex, p);
	    }

	    m = p.matcher(input);
	}

	if (!m.matches()) return null;

	LinkedList<String> list = new LinkedList<String>();

	for(int i=1;i < m.groupCount()+1;i++) {
	    String g = m.group(i);
	    if (g == null) break;
	    list.add(g);
	}

	return list.toArray(new String[0]);
    }

    public static HashMap<String, String[]> analyseCommandLine(String[] args, HashMap<String, Integer> options) {
	if (args.length == 0) return null;

	HashMap<String, String[]> ret = new HashMap<String, String[]>();

	for(int i=0;i<args.length;i++) {
	    boolean found = false;
	    for(String s : options.keySet()) {
		if (s.equals(args[i])) {
		    int narg = options.get(s);
		    if (args.length - i >= narg + 1) {
			String[] a = new String[narg];
			for(int j=0;j<narg;j++) {
			    a[j] = args[i + 1 + j];
			}
			ret.put(s, a);
			i += narg;
			found = true;
			break;
		    }
		}
	    }
	    if (!found) return null;
	}

	return ret;
    }
}
