package org.naokishibata.armarker;

import java.io.*;
import java.awt.*;
import javax.swing.*;

import org.naokishibata.generic.*;

/**
 * This is a class for drawing an AccuTag marker. The marker is
 * displayed on the screen while output to "marker.png" file.
 */
public class DrawMarker3 {
    static int[] bittab = {
	12,  0, 13,  1, -1,
	02, 14,  3, 15,  4,
	16,  5, -2,  6, 17,
	07, 18,  8, 19,  9,
	-2, 10, 20, 11, -1,
    };

    static final int POLY = 0xd5;

    static int crc9(int r) {
	r = 0xfff & (r ^ 0x555);
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	r = (r << 1) ^ (POLY & -((r >> 11) & 1));
	return 0x1ff & r;
    }

    DrawMarker3() {}

    /**
     * The main function. You can specify 12 bits of payload value as
     * the command line argument.
     */
    public static void main(String[] args) throws Exception {
	int payload0 = 0;

	if (args.length != 0) {
	    payload0 = Integer.parseInt(args[0]);
	}

	int payload = (payload0 & 0xfff);
	payload = (payload & 0xfff);
	payload |= (crc9(payload) << 12);

	int bs = 64;

	Scribble2D s = new Scribble2D(bs*9, bs*9);

	JFrame frame = s.makeFrame();

	frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

	int xoffset = 0;
	int yoffset = 0;

	int c = 0;
	Color col = new Color(c, c, c);

	for(int y=0;y<9;y++) {
	    for(int x=0;x<9;x++) {
		boolean b = false;

		if (y == 0) continue;
		if (y == 8) continue;
		if (x == 0) continue;
		if (x == 8) continue;

		if (y == 1) b = true;
		if (y == 7) b = true;
		if (x == 1) b = true;
		if (x == 7) b = true;

		if (2 <= y && y <= 6 && 2 <= x && x <= 6) {
		    int t = bittab[(y-2) * 5 + (x-2)];
		    if (t == -1) {
			b = true;
		    } else if (t == -2) {
			b = false;
		    } else {
			b = ((payload >> t) & 1) != 0;
		    }
		}

		if (b) {
		    s.fillBox(xoffset + bs*x, yoffset + bs*y, xoffset + bs*(x+1), yoffset + bs*(y+1), col);
		}
	    }
	}

	s.fillLines(col, bs*2, bs*1, bs*1, bs*2, bs*2, bs*2);

	s.saveImage("png", new File("marker.png"));

	System.out.println();
	System.out.println("The marker image with payload '" + payload0 + "' is saved to marker.png");
	if (args.length == 0) {
	    System.out.println("You can specify a 12bit payload from the command line argument");
	}
	System.out.println("Close the window to exit");
    }
}
