package org.naokishibata.examples;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.nio.*;

import com.sun.opengl.util.*;
import javax.media.opengl.*;

import org.naokishibata.generic.*;
import org.naokishibata.ocvimgio.*;

public class CVGLTest implements GLEventListener {
    static final boolean debug = true;
    static final int MAPSIZE = 1024;

    private GLCanvas canvas;
    private int texName;
    double tw, th;

    CvCapture capture;
    int iw, ih;
    ByteBuffer bb = ByteBuffer.allocateDirect(MAPSIZE * MAPSIZE * 3);

    private boolean firstShow = true;

    public CVGLTest() {
	capture = CvCapture.cvCreateCameraCapture(0);
	capture.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_WIDTH , 1024);
	capture.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_HEIGHT,  768);

	IplImage iimg = capture.cvQueryFrame();

	iw = iimg.width();
	ih = iimg.height();

	tw = (double)iw / MAPSIZE;
	th = (double)ih / MAPSIZE;

	//

        GLCapabilities caps = new GLCapabilities();

	caps.setAlphaBits(8);
	caps.setDepthBits(24);

        caps.setDoubleBuffered(true);
        caps.setHardwareAccelerated(true);

        canvas = new GLCanvas(caps, null, null, null);

        canvas.addGLEventListener(this);
    }

    public void init(GLAutoDrawable drawable) {
	if (firstShow) {
	    firstShow = false;
	    if (debug) drawable.setGL(new DebugGL(drawable.getGL()));

	    GL gl = drawable.getGL();

	    gl.glMatrixMode(GL.GL_PROJECTION);
	    gl.glLoadIdentity();
	    gl.glOrtho(0, 1, 0, 1, -1, 1);

	    //

	    int[] tmp = new int[1];
	    gl.glGenTextures(1, tmp, 0);
	    texName = tmp[0];

	    gl.glBindTexture(GL.GL_TEXTURE_2D, texName);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_GENERATE_MIPMAP, GL.GL_FALSE);
	    gl.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA8, MAPSIZE, MAPSIZE, 0, GL.GL_BGR, GL.GL_UNSIGNED_BYTE, bb.rewind());

	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
	    gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_REPLACE);
	}
    }

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
	double height0 = (double)height / ih * iw;
	double size = width < height0 ? width : height0;
	double hpad = (double)(width   - size) / size / 2;
	double vpad = (double)(height0 - size) / size / 2;

	GL gl = drawable.getGL();

	gl.glMatrixMode(GL.GL_PROJECTION);
	gl.glLoadIdentity();
	gl.glFrustum(-hpad, 1 + hpad, -vpad, 1 + vpad,  1, 10000);
    }

    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
    }

    public void display(GLAutoDrawable drawable) {
	IplImage iimg = capture.cvRetrieveFrame();
	capture.cvGrabFrame();

	GL gl = drawable.getGL();

	gl.glBindTexture(GL.GL_TEXTURE_2D, texName);

	gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_GENERATE_MIPMAP, GL.GL_FALSE);

	gl.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, 0, 0, iw, ih, GL.GL_BGR, GL.GL_UNSIGNED_BYTE, iimg.getDirect().rewind());

	gl.glEnable(GL.GL_TEXTURE_2D);

	gl.glBegin( GL.GL_QUADS );
	gl.glTexCoord2d(0.0,th ); gl.glVertex3d(0.0,0.0, -1.0);
	gl.glTexCoord2d(tw ,th ); gl.glVertex3d(1.0,0.0, -1.0);
	gl.glTexCoord2d(tw ,0.0); gl.glVertex3d(1.0,1.0, -1.0);
	gl.glTexCoord2d(0.0,0.0); gl.glVertex3d(0.0,1.0, -1.0);
	gl.glEnd();
    }

    public static void main(String[] args) {
	JFrame frame = new JFrame();

	frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        frame.setLayout(new BorderLayout());

        frame.setSize(1024, 768);

	AWTUtils.placeAtTheCenter(frame);

	CVGLTest instance = new CVGLTest();

        frame.add(instance.canvas, BorderLayout.CENTER);

        frame.setVisible(true);

	FPSAnimator anim = new FPSAnimator(instance.canvas, 60);
	anim.start();
    }
}
