package org.naokishibata.generic;

import java.util.*;

public class LRUMap<K,V> extends HashMap<K,V> {
    private int capacity;
    private LinkedHashSet<V> set = new LinkedHashSet<V>();
    private HashMap<V, HashSet<K>> revMap = new HashMap<V, HashSet<K>>();

    public LRUMap(int capacity) {
	super();
	this.capacity = capacity;
    }

    public void clear() {
	super.clear();
	set.clear();
    }

    private void enforceCapacity() {
	while (set.size() > capacity) {
	    V val = MiscUtils.getFirst(set);
	    HashSet<K> keys = revMap.get(val);

	    for(K k : keys) super.remove(k);
	    set.remove(val);
	    revMap.remove(val);
	}
    }

    public void setCapacity(int c) {
	capacity = c;

	enforceCapacity();
    }

    public void renewKey(K key) {
	if (!containsKey(key)) return;
	V val = super.get(key);
	set.remove(val);
	set.add(val);
    }

    public void renewValue(V val) {
	if (set.contains(val)) {
	    set.remove(val);
	    set.add(val);
	}
    }

    public V put(K key, V value) {
	V ret = super.put(key, value);
	set.remove(value);
	set.add(value);

	HashSet<K> rmSet = revMap.get(value);
	if (rmSet == null) {
	    rmSet = new HashSet<K>();
	    revMap.put(value, rmSet);
	}
	rmSet.add(key);

	enforceCapacity();

	return ret;
    }

    public void putAll(Map<? extends K, ? extends V> m) {
	for(K key : m.keySet()) {
	    V val = m.get(key);
	    put(key,val);
	}
    }

    public V remove(Object key) {
	V val = super.remove(key);
	revMap.remove(val);
	set.remove(val);

	return val;
    }
}
