package org.naokishibata.generic;

import java.awt.*;
import java.awt.event.*;

public class MouseSensor implements MouseMotionListener, MouseListener, MouseWheelListener {
    public final static int LEFT   = 1 << 0;
    public final static int MIDDLE = 1 << 1;
    public final static int RIGHT  = 1 << 2;

    private int mouseX = 0, mouseY = 0, buttonState = 0, wheelPos = 0;

    public int posX() { return mouseX; }
    public int posY() { return mouseY; }
    public int button() { return buttonState; }
    public int wheel() { return wheelPos; }

    public void resetWheel() { wheelPos = 0; }

    public Component component;

    public void installOn(Component c) {
	buttonState = 0;
	wheelPos = 0;

	c.addMouseListener(this);
	c.addMouseMotionListener(this);
	c.addMouseWheelListener(this);

	component = c;
    }

    public void uninstallFrom(Component c) {
	c.removeMouseListener(this);
	c.removeMouseMotionListener(this);
	c.removeMouseWheelListener(this);

	component = null;
    }

    public Component component() { return component; }

    //

    public void mouseMoved(MouseEvent e) {
	mouseX = e.getX();
	mouseY = e.getY();
    }

    public void mouseDragged(MouseEvent e) {
	mouseX = e.getX();
	mouseY = e.getY();
    }

    public void mousePressed(MouseEvent e) {
	int m = e.getModifiers();
	if ((m & InputEvent.BUTTON1_MASK) != 0) buttonState |= LEFT;
	if ((m & InputEvent.BUTTON2_MASK) != 0) buttonState |= MIDDLE;
	if ((m & InputEvent.BUTTON3_MASK) != 0) buttonState |= RIGHT;
    }

    public void mouseReleased(MouseEvent e) {
	int m = e.getModifiers();
	if ((m & InputEvent.BUTTON1_MASK) != 0) buttonState &= ~LEFT;
	if ((m & InputEvent.BUTTON2_MASK) != 0) buttonState &= ~MIDDLE;
	if ((m & InputEvent.BUTTON3_MASK) != 0) buttonState &= ~RIGHT;
    }

    public void mouseWheelMoved(MouseWheelEvent e) {
	int notches = e.getWheelRotation();

	if (notches < 0) {
	    wheelPos--;
	} else {
	    wheelPos++;
	}
    }

    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) {}
}
