package org.naokishibata.generic;

import javax.swing.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

public class StringInputDialog implements ActionListener {
    public static final int OK_OPTION = JOptionPane.OK_OPTION;
    public static final int OK_CANCEL_OPTION = JOptionPane.OK_CANCEL_OPTION;

    JDialog dialog;
    JButton butOK, butCancel;
    RegexTextField text;
    boolean OKPressed;
    final int butWidth = 80;

    public StringInputDialog(Frame frame, String message, String title, String patDuringInput, String patAfterInput, int optionType) {
	dialog = new JDialog(frame, title, true);
	JPanel panel = new JPanel();
	dialog.setContentPane(panel);

	FormulaLayout fl = new FormulaLayout(320, 170, 240, 170);

	panel.setLayout(fl);

	dialog.setSize(320, 170);
	dialog.setMinimumSize(new Dimension(240,170));
	dialog.setResizable(true);

	//

	if (optionType == OK_OPTION) {
	    butOK = new JButton("OK");
	    butOK.addActionListener(this);
	    butOK.setMaximumSize(new Dimension(butWidth, butOK.getPreferredSize().height));
	    butOK.setMinimumSize(new Dimension(butWidth, butOK.getPreferredSize().height));
	    panel.add(butOK);
	    dialog.getRootPane().setDefaultButton(butOK);

	    fl.setFormula(butOK, new FormulaLayout.LayoutFormula() {
		    public Rectangle getLayout(int w, int h) {
			Dimension d = butOK.getPreferredSize();
			return new Rectangle(w - butWidth - 10, h - d.height - 10, butWidth, d.height);
		    }
		});
	} else {
	    butCancel = new JButton("Cancel");
	    butCancel.addActionListener(this);
	    butCancel.setMaximumSize(new Dimension(butWidth, butCancel.getPreferredSize().height));
	    butCancel.setMinimumSize(new Dimension(butWidth, butCancel.getPreferredSize().height));
	    //butCancel.setLocation(10,10);
	    panel.add(butCancel);

	    fl.setFormula(butCancel, new FormulaLayout.LayoutFormula() {
		    public Rectangle getLayout(int w, int h) {
			Dimension d = butCancel.getPreferredSize();
			return new Rectangle(w - butWidth - 10, h - d.height - 10, butWidth, d.height);
		    }
		});

	    butOK = new JButton("OK");
	    butOK.addActionListener(this);
	    butOK.setMaximumSize(new Dimension(butWidth, butOK.getPreferredSize().height));
	    butOK.setMinimumSize(new Dimension(butWidth, butOK.getPreferredSize().height));
	    panel.add(butOK);
	    dialog.getRootPane().setDefaultButton(butOK);

	    fl.setFormula(butOK, new FormulaLayout.LayoutFormula() {
		    public Rectangle getLayout(int w, int h) {
			Dimension d = butOK.getPreferredSize();
			return new Rectangle(w - butWidth*2 - 20, h - d.height - 10, butWidth, d.height);
		    }
		});
	}

	final JLabel label = new JLabel(message);
	panel.add(label);

	label.setFont(new Font("SansSerif", Font.PLAIN, 16));

	fl.setFormula(label, new FormulaLayout.LayoutFormula() {
		public Rectangle getLayout(int w, int h) {
		    Dimension d = label.getPreferredSize();
		    return new Rectangle(20, 20, d.width, d.height);
		}
	    });

	text = new RegexTextField(patDuringInput, patAfterInput, 8);

	text.setFont(new Font("SansSerif", Font.PLAIN, 16));
	text.setSelectWhenFocused(true);

	panel.add(text);

	fl.setFormula(text, new FormulaLayout.LayoutFormula() {
		public Rectangle getLayout(int w, int h) {
		    Dimension d = text.getPreferredSize();
		    return new Rectangle(20, 60, w - 40, d.height);
		}
	    });
    }

    public void actionPerformed(ActionEvent e)
    {
	Object o = e.getSource();
	if (!(o instanceof Component)) return;

	Component c = (Component)e.getSource();

	if (text.isAcceptable()) {
	    if (c == butOK) {
		OKPressed = true;
		dialog.setVisible(false);
	    } else if (c == butCancel) {
		dialog.setVisible(false);
	    }
	}
    }

    public boolean showDialog(String init) {
	text.setText(init);
	OKPressed = false;

	AWTUtils.placeAtTheCenter(dialog);
	dialog.setVisible(true);

	return OKPressed;
    }

    public String getText() { return text.getText(); }

    public static Double showDialogForDoubleInput(Frame owner, String message, String title, int optionType, double initVal) {
	StringInputDialog d = new StringInputDialog(owner, message, title,
						    "[-+]?[0-9]*\\.?(([0-9]*)|([0-9]+((([eE][-+]?)|([eE][-+]?[0-9]+)))?))",
						    "[-+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?", optionType);
	d.showDialog("" + initVal);

	try {
	    if (d.OKPressed) return new Double(Double.parseDouble(d.getText()));
	} catch(NumberFormatException e) {
	}

	return null;
    }

    public static Integer showDialogForIntInput(Frame owner, String message, String title, int optionType, int initVal) {
	StringInputDialog d = new StringInputDialog(owner, message, title, "[-+]?[0-9]*", "[-+]?[0-9]+", optionType);

	d.showDialog("" + initVal);

	try {
	    if (d.OKPressed) return new Integer(Integer.parseInt(d.getText()));
	} catch(NumberFormatException e) {
	}

	return null;
    }

    public static String showDialogForStringInput(Frame owner, String message, String title, int optionType, String initText) {
	StringInputDialog d = new StringInputDialog(owner, message, title, ".*", ".*", optionType);

	d.showDialog(initText);

	if (d.OKPressed) return d.getText();

	return null;
    }

    public static void main(String[] args) {
	System.out.println(showDialogForDoubleInput(null, "input a double number", "input", OK_CANCEL_OPTION, 3.1415));
	System.exit(0);
    }
}
