package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

public class LineSegment3D implements Serializable {
    //private static final long serialVersionUID = MiscUtils.svuid();

    private final Vector3D u[];

    private void check() {
	assert u[0] != null && u[1] != null;
	assert !Vector3D.epsilonEquals(u[0],u[1]);
    }

    public LineSegment3D(Vector3D v1, Vector3D v2) {
	u = new Vector3D[2];
	u[0] = v1; u[1] = v2;

	//if (Debug.enableDebug && Debug.checkLevel >= 10) check();
    }

    public Vector3D v(int i) { return u[i]; }

    public String toString() {
	return "(LineSegment "+u[0].toString()+"-"+u[1].toString()+")";
    }

    //

    public static double length(LineSegment3D ls) {
	return Vector3D.distance(ls.u[0],ls.u[1]);
    }

    public static double distance(LineSegment3D l, Vector3D v) {
	boolean b0 = Vector3D.dotProduct(Vector3D.sub(v,l.u[0]),Vector3D.sub(l.u[1],l.u[0])) >= 0;
	boolean b1 = Vector3D.dotProduct(Vector3D.sub(v,l.u[1]),Vector3D.sub(l.u[0],l.u[1])) >= 0;

	if (b0 && b1) return Line3D.distance(new Line3D(l), v);
	if (b0) return Vector3D.distance(l.u[1],v); else return Vector3D.distance(l.u[0],v);
    }

    public static int includes(LineSegment3D l, Vector3D v, double eps) {
	if (Vector3D.distance(l.v(0),v) <= eps) return 1;
	if (Vector3D.distance(l.v(1),v) <= eps) return 1;

	if (Line3D.distance(new Line3D(l), v) > eps) return 0;
	if (v.minus(l.v(0)).dot(v.minus(l.v(1))) > 0) return 0;
	return 2;

	//return distance(l, v) <= eps ? 2 : 0;
    }

    public static int includes(LineSegment3D l, Vector3D v) {
	return includes(l,v,MathUtils.epsilon);
    }

    /*
    public static double distance(LineSegment3D m, Line3D l) {
	LineSegment3D ls = Line3D.nearestPointsBetween(new Line3D(m), l);

	if (includes(m, ls.v(0)) != 0) {
	    return LineSegment3D.length(ls);
	}

	return Math.min(Line3D.distance(l,m.u[0]),Line3D.distance(l,m.u[1]));
    }

    public static int intersects(LineSegment3D m, LineSegment3D l) {
	LineSegment3D ls = Line3D.nearestPointsBetween(new Line3D(m), new Line3D(l));

	int i0 = includes(m, ls.v(1));
	int i1 = includes(l, ls.v(0));

	if (i0 == 2 && i1 == 2) return 2;
	if (i0 != 0 && i1 != 0) return 1;
	return 0;
    }

    public static LineSegment3D nearestPointsBetween(LineSegment3D ls0, Line3D l) {
	LineSegment3D ls1 = Line3D.nearestPointsBetween(new Line3D(ls0), l);

	if (includes(ls0,ls1.u[0]) != 0) return ls1;

	if (Line3D.distance(l,ls0.u[0]) < Line3D.distance(l,ls0.u[1])) {
	    return new LineSegment3D(ls0.u[0], ls1.u[1]);
	} else {
	    return new LineSegment3D(ls0.u[1], ls1.u[1]);
	}
    }

    public static LineSegment3D nearestPointsBetween(LineSegment3D ls0, LineSegment3D ls1) {
	LineSegment3D np = Line3D.nearestPointsBetween(new Line3D(ls0), new Line3D(ls1));
	boolean b0 = includes(ls0, np.v(0)) != 0;
	boolean b1 = includes(ls1, np.v(1)) != 0;
	
	if (b0 && b1) return np;

	if (b0) {
	    if (Line3D.distance(new Line3D(ls0),ls1.u[0]) < Line3D.distance(new Line3D(ls0),ls1.u[1])) {
		return new LineSegment3D(np.v(0), ls1.u[0]);
	    } else {
		return new LineSegment3D(np.v(0), ls1.u[1]);
	    }
	}

	if (b1) {
	    if (Line3D.distance(new Line3D(ls1),ls0.u[0]) < Line3D.distance(new Line3D(ls1),ls0.u[1])) {
		return new LineSegment3D(ls0.u[0], np.v(1));
	    } else {
		return new LineSegment3D(ls0.u[1], np.v(1));
	    }
	}

	double d00 = Vector3D.distance(ls0.u[0],ls1.u[0]);
	double d01 = Vector3D.distance(ls0.u[0],ls1.u[1]);
	double d10 = Vector3D.distance(ls0.u[1],ls1.u[0]);
	double d11 = Vector3D.distance(ls0.u[1],ls1.u[1]);

	if (d00 <= d01 && d00 <= d10 && d00 <= d11) return new LineSegment3D(ls0.u[0], ls1.u[0]);

	if (d01 <= d10 && d01 <= d11) return new LineSegment3D(ls0.u[0], ls1.u[1]);

	if (d10 <= d11) return new LineSegment3D(ls0.u[0], ls1.u[0]);

	return new LineSegment3D(ls0.u[1], ls1.u[1]);
    }
    */

    public static Vector3D midpoint(LineSegment3D ls) {
	return Vector3D.midpoint(ls.v(0),ls.v(1));
    }

    public Vector3D midpoint() {
	return Vector3D.midpoint(v(0),v(1));
    }

    public boolean epsilonEquals(LineSegment3D ls, double eps) {
	if (v(0).epsilonEquals(ls.v(0), eps) &&
	    v(1).epsilonEquals(ls.v(1), eps)    ) return true;

	if (v(0).epsilonEquals(ls.v(1), eps) &&
	    v(1).epsilonEquals(ls.v(0), eps)    ) return true;

	return false;
    }

    public boolean epsilonEquals(LineSegment3D ls) {
	return epsilonEquals(ls, MathUtils.epsilon);
    }


    public double distanceFrom(Vector3D v) {
	return distance(this,v);
    }

    public int includes(Vector3D v) {
	return includes(this, v);
    }

    public double length() { return length(this);}

    /*
    public double distance(LineSegment3D l) {
	return length(nearestPointsBetween(this,l));
    }

    public Vector3D nearestPoint(Vector3D v) {
	Vector3D p = Line3D.nearestPoint(new Line3D(this),v);

	if (includes(this,p) != 0) return p;

	if (Vector3D.distance(u[0],v) < Vector3D.distance(u[1],v)) return u[0]; else return u[1];
    }

    public boolean includes(LineSegment3D ls) {
	Vector3D v = Line3D.nearestPointsBetween(new Line3D(this), new Line3D(ls)).v(0);

	return includes(this,v) != 0 && includes(ls,v) != 0;
    }
    */

    public int hashCode() {
	//return Vector3D.midpoint(u[0], u[1]).hashCode();

	return (int)(
		     Double.doubleToRawLongBits(u[0].x() + u[1].x()) ^
		     Double.doubleToRawLongBits(u[0].y() + u[1].y()) ^
		     Double.doubleToRawLongBits(u[0].z() + u[1].z())
		     );
    }

    public boolean equals(Object o) {
	if (!(o instanceof LineSegment3D)) return false;
	LineSegment3D ls = (LineSegment3D) o;

	if (u == null || ls.u == null) return false;  // serialization bug work around

	if (Vector3D.equals(u[0],ls.u[0]) && Vector3D.equals(u[1],ls.u[1])) return true;
	if (Vector3D.equals(u[0],ls.u[1]) && Vector3D.equals(u[1],ls.u[0])) return true;

	return false;
    }
}
